@@ -21,7 +21,9 @@ public:
   }
   SEXP capture() {
     SEXP raster = PROTECT(Rf_allocVector(INTSXP, this->width * this->height));
-    memcpy(INTEGER(raster), this->buffer, this->width * this->height * 4);
+    agg::rendering_buffer caprbuf(reinterpret_cast<agg::int8u*>(INTEGER(raster)),
+                                  this->width, this->height, this->width * 4);
+    agg::convert<pixfmt_r_capture, pixfmt_type_32>(&caprbuf, &this->rbuf);
     SEXP dims = PROTECT(Rf_allocVector(INTSXP, 2));
     INTEGER(dims)[0] = this->height;
     INTEGER(dims)[1] = this->width;
